"""Tests for the requests module."""

import pytest

from pysmartapp.request import EmptyDataResponse, Request, Response

from .utilities import get_fixture


class TestResponse:
    """Tests for the Response class."""

    @staticmethod
    def test_to_data():
        """Tests the to_data method."""
        resp = Response()
        with pytest.raises(NotImplementedError):
            resp.to_data()


class TestEmptyDataResponse:
    """Tests for the EmptyDataResponse class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        resp = EmptyDataResponse('tag')
        assert resp.name == 'tag'

    @staticmethod
    def test_to_data():
        """Tests the to_data method."""
        resp = EmptyDataResponse('tag')
        result = resp.to_data()
        assert result == {'tag': {}}

    @staticmethod
    def test_name():
        """Tests the name setter."""
        resp = EmptyDataResponse('tag')
        resp.name = 'tag2'
        assert resp.name == 'tag2'


class TestRequest:
    """Tests for the Request class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_process_raises():
        """Tests the to_data method."""
        # Arrange
        data = get_fixture('ping_request')
        resp = Request(data)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await resp.process(None, validate_signature=False)



""""Tests for the oauthcallback module."""

from pysmartapp.const import LIFECYCLE_OAUTH_CALLBACK
from pysmartapp.oauthcallback import OAuthCallbackRequest

from .utilities import get_fixture


class TestOAuthCallbackRequest:
    """Tests for the OAuthCallbackRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('oauth_callback_request')
        # Act
        req = OAuthCallbackRequest(data)
        # Assert
        assert req.oauth_callback_data_raw == data['oAuthCallbackData']
        assert req.lifecycle == LIFECYCLE_OAUTH_CALLBACK
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == '8a0dcdc9-1ab4-4c60-9de7-cb78f59a1121'
        assert req.url_path == 'string'



"""Define the ping module."""

from .request import Request, Response


class PingRequest(Request):
    """Defines a ping request."""

    def __init__(self, data: dict):
        """Create a new instance of the PingRequest class."""
        super().__init__(data)
        self._supports_validation = False
        self._ping_data_raw = data['pingData']

    async def _process(self, app):
        response = PingResponse()
        response.ping_challenge = self.ping_challenge
        return response

    @property
    def ping_data_raw(self) -> dict:
        """Get the raw data structure of the ping request."""
        return self._ping_data_raw

    @property
    def ping_challenge(self):
        """Get the challenge code as part of the request."""
        return self._ping_data_raw['challenge']


class PingResponse(Response):
    """Defines a ping response."""

    def __init__(self):
        """Create a new instance of the PingResponse."""
        self._ping_challenge = None

    def to_data(self) -> dict:
        """Create a data structure for the response."""
        return {'pingData': {'challenge': self.ping_challenge}}

    @property
    def ping_challenge(self):
        """Get the ping challenge in the response."""
        return self._ping_challenge

    @ping_challenge.setter
    def ping_challenge(self, value: str):
        """Set the ping challenge response."""
        self._ping_challenge = value



"""Define common test configuraiton."""

import pytest

from pysmartapp.dispatch import Dispatcher
from pysmartapp.smartapp import SmartApp, SmartAppManager


@pytest.fixture
def smartapp(event_loop) -> SmartApp:
    """Fixture for testing against the SmartApp class."""
    app = SmartApp(dispatcher=Dispatcher(loop=event_loop))
    app.name = 'SmartApp'
    app.description = 'SmartApp Description'
    app.permissions.append('l:devices')
    app.config_app_id = 'myapp'
    return app


@pytest.fixture
def manager(event_loop) -> SmartAppManager:
    """Fixture for testing against the SmartAppManager class."""
    return SmartAppManager('/path/to/app',
                           dispatcher=Dispatcher(loop=event_loop))


@pytest.fixture
def handler():
    """Fixture handler to mock in the dispatcher."""
    def target(*args, **kwargs):
        target.fired = True
        target.args = args
        target.kwargs = kwargs
    target.fired = False
    return target


@pytest.fixture
def async_handler():
    """Fixture async handler to mock in the dispatcher."""
    async def target(*args, **kwargs):
        target.fired = True
        target.args = args
        target.kwargs = kwargs
    target.fired = False
    return target



"""Define the uninstall module."""

from .request import EmptyDataResponse, Request, Response


class UninstallRequest(Request):
    """Define the UninstallRequest class."""

    def __init__(self, data: dict):
        """Create a new instance of the UninstallRequest."""
        super().__init__(data)
        uninstall_data = self._uninstall_data_raw = data['uninstallData']
        self._init_installed_app(uninstall_data['installedApp'])

    async def _process(self, app) -> Response:
        resp = EmptyDataResponse('uninstallData')
        return resp

    @property
    def uninstall_data_raw(self) -> dict:
        """Get the raw update data."""
        return self._uninstall_data_raw



""""Tests for the config module."""

import pytest

from pysmartapp.config import ConfigRequest
from pysmartapp.const import LIFECYCLE_CONFIG, LIFECYCLE_CONFIG_INIT

from .utilities import get_fixture


class TestConfigRequest:
    """Tests for the ConfigRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('config_init_request')
        # Act
        req = ConfigRequest(data)
        # Assert
        assert req.config_data_raw == data['configurationData']
        assert req.lifecycle == LIFECYCLE_CONFIG
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == \
            data['configurationData']['installedAppId']
        assert req.phase == LIFECYCLE_CONFIG_INIT
        assert req.page_id == ''
        assert req.previous_page_id == ''

    @staticmethod
    @pytest.mark.asyncio
    async def test_process_invalid():
        """Tests the process method for an invalid phase"""
        # Arrange
        data = get_fixture('config_init_request')
        data['configurationData']['phase'] = "UNKNOWN"
        # Act
        req = ConfigRequest(data)
        # Assert
        with pytest.raises(ValueError):
            await req.process(None, validate_signature=False)



"""Defines the dispatch component for notifying others of signals."""

import asyncio
from collections import defaultdict
from typing import Any, Callable, Dict, List, Sequence

TargetType = Callable[..., Any]
DisconnectType = Callable[[], None]
ConnectType = Callable[[str, TargetType], DisconnectType]
SendType = Callable[..., Sequence[asyncio.Future]]


class Dispatcher:
    """Define the dispatch class."""

    def __init__(self, *, connect: ConnectType = None, send: SendType = None,
                 signal_prefix: str = '', loop=None):
        """Create a new instance of the dispatch component."""
        self._signal_prefix = signal_prefix
        self._signals = defaultdict(list)
        self._loop = loop or asyncio.get_event_loop()
        self._connect = connect or self._default_connect
        self._send = send or self._default_send
        self._last_sent = []

    def connect(self, signal: str, target: TargetType) \
            -> DisconnectType:
        """Connect function to signal.  Must be ran in the event loop."""
        return self._connect(self._signal_prefix + signal, target)

    def send(self, signal: str, *args: Any) -> Sequence[asyncio.Future]:
        """Fire a signal.  Must be ran in the event loop."""
        sent = self._last_sent = self._send(
            self._signal_prefix + signal, *args)
        return sent

    def _default_connect(self, signal: str, target: TargetType) \
            -> DisconnectType:
        """Connect function to signal.  Must be ran in the event loop."""
        self._signals[signal].append(target)

        def remove_dispatcher() -> None:
            """Remove signal listener."""
            try:
                self._signals[signal].remove(target)
            except ValueError:
                # signal was already removed
                pass
        return remove_dispatcher

    def _default_send(self, signal: str, *args: Any) -> \
            Sequence[asyncio.Future]:
        """Fire a signal.  Must be ran in the event loop."""
        targets = self._signals[signal]
        futures = []
        for target in targets:
            task = self._call_target(target, *args)
            futures.append(task)
        return futures

    def _call_target(self, target, *args) -> asyncio.Future:
        if asyncio.iscoroutinefunction(target):
            return self._loop.create_task(target(*args))
        return self._loop.run_in_executor(None, target, *args)

    @property
    def signals(self) -> Dict[str, List[TargetType]]:
        """Get the dictionary of registered signals and callbaks."""
        return self._signals

    @property
    def last_sent(self) -> Sequence[asyncio.Future]:
        """Get the last sent asyncio tasks."""
        return self._last_sent



"""Tests for the SmartApp file."""

import asyncio

import pytest

from pysmartapp.dispatch import Dispatcher
from pysmartapp.errors import (
    SignatureVerificationError, SmartAppNotRegisteredError)
from pysmartapp.smartapp import SmartApp, SmartAppManager

from .utilities import get_dispatch_handler, get_fixture

INSTALLED_APP_ID = '8a0dcdc9-1ab4-4c60-9de7-cb78f59a1121'
APP_ID = 'f6c071aa-6ae7-463f-b0ad-8620ac23140f'


class TestSmartApp:
    """Tests for the SmartApp class."""

    @staticmethod
    def test_initialize():
        """Tests the property initialization."""
        # Arrange
        path = '/my/test/path'
        public_key = 'test'
        dispatcher = Dispatcher()
        # Act
        app = SmartApp(path=path, public_key=public_key,
                       dispatcher=dispatcher)
        # Assert
        assert app.path == path
        assert app.public_key == public_key
        assert app.dispatcher == dispatcher
        assert app.permissions == []
        assert app.config_app_id == 'app'

    @staticmethod
    def test_setters():
        """Tests the property setters."""
        # Arrange
        app = SmartApp()
        # Act
        app.app_id = "Test"
        app.config_app_id = "Test Config Id"
        app.description = "Description"
        app.name = "Name"
        # Assert
        assert app.app_id == "Test"
        assert app.config_app_id == "Test Config Id"
        assert app.description == "Description"
        assert app.name == "Name"

    @staticmethod
    @pytest.mark.asyncio
    async def test_ping(smartapp):
        """Tests the ping lifecycle event."""
        # Arrange
        request = get_fixture("ping_request")
        expected_response = get_fixture("ping_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_ping(handler)
        # Act
        response = await smartapp.handle_request(request)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_config_init(smartapp):
        """Tests the configuration initialization lifecycle event."""
        # Arrange
        request = get_fixture("config_init_request")
        expected_response = get_fixture("config_init_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_config(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_config_page(smartapp):
        """Tests the configuration initialization page event."""
        # Arrange
        request = get_fixture("config_page_request")
        expected_response = get_fixture("config_page_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_config(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_install(smartapp):
        """Tests the install lifecycle event."""
        # Arrange
        request = get_fixture("install_request")
        expected_response = get_fixture("install_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_install(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_update(smartapp):
        """Tests the update lifecycle event."""
        # Arrange
        request = get_fixture("update_request")
        expected_response = get_fixture("update_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_update(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_event(smartapp):
        """Tests the event lifecycle event."""
        # Arrange
        request = get_fixture("event_request")
        expected_response = get_fixture("event_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_event(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_oauth_callback(smartapp):
        """Tests the oauth_callback lifecycle event."""
        # Arrange
        request = get_fixture("oauth_callback_request")
        expected_response = get_fixture("oauth_callback_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_oauth_callback(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_uninstall(smartapp):
        """Tests the uninstall lifecycle event."""
        # Arrange
        request = get_fixture("uninstall_request")
        expected_response = get_fixture("uninstall_response")
        handler = get_dispatch_handler(smartapp)
        smartapp.connect_uninstall(handler)
        # Act
        response = await smartapp.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*smartapp.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_sig_verification():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_pass_request')
        smartapp = SmartApp(public_key=public_key)
        # Act
        resp = await smartapp.handle_request(
            data['body'], data['headers'], True)
        # Assert
        assert resp

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_sig_verification_missing_headers():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_pass_request')
        smartapp = SmartApp(public_key=public_key)
        # Act/Assert
        with pytest.raises(SignatureVerificationError):
            await smartapp.handle_request(data['body'], [], True)

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_sig_verification_fails():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_fail_request')
        smartapp = SmartApp(public_key=public_key)
        # Act/Assert
        with pytest.raises(SignatureVerificationError):
            await smartapp.handle_request(data['body'], data['headers'], True)


class TestSmartAppManager:
    """Tests for the SmartAppManager class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_ping_not_registered(manager):
        """Tests the ping lifecycle event with no registered apps."""
        # Arrange
        request = get_fixture("ping_request")
        expected_response = get_fixture("ping_response")
        handler = get_dispatch_handler(manager)
        manager.connect_ping(handler)
        # Act
        response = await manager.handle_request(request)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired
        assert response == expected_response

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_not_registered(manager: SmartAppManager):
        """Tests processing a request when no SmartApp has been registered."""
        # Arrange
        request = get_fixture("config_init_request")
        # Act
        with pytest.raises(SmartAppNotRegisteredError) as e_info:
            await manager.handle_request(request, None, False)
        # Assert
        assert e_info.value.installed_app_id == INSTALLED_APP_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_handle_request_not_app_id(manager: SmartAppManager):
        """Tests processing a request when no SmartApp has been registered."""
        # Arrange
        request = get_fixture("config_init_sig_fail_request")['body']
        # Act
        with pytest.raises(SmartAppNotRegisteredError) as e_info:
            await manager.handle_request(request, None, False)
        # Assert
        assert e_info.value.installed_app_id == INSTALLED_APP_ID

    @staticmethod
    def test_register(manager: SmartAppManager):
        """Test register"""
        public_key = '123'
        # Act
        app = manager.register(APP_ID, public_key)
        # Assert
        assert app.app_id == APP_ID
        assert app.public_key == public_key
        assert app.path == manager.path
        assert APP_ID in manager.smartapps

    @staticmethod
    def test_register_no_app_id(manager: SmartAppManager):
        """Test register with no SmartApp app id"""
        # Act
        with pytest.raises(ValueError) as e_info:
            manager.register(None, '')
        # Assert
        assert str(e_info.value) == 'smartapp must have an app_id.'

    @staticmethod
    def test_register_twice(manager: SmartAppManager):
        """Test register with the same app twice"""
        # Arrange
        public_key = '123'
        manager.register(APP_ID, public_key)
        # Act
        with pytest.raises(ValueError) as e_info:
            manager.register(APP_ID, public_key)
        # Assert
        assert str(e_info.value) == 'smartapp already registered.'

    @staticmethod
    def test_unregister(manager: SmartAppManager):
        """Test unregister"""
        # Arrange'
        manager.register(APP_ID, '123')
        # Act
        manager.unregister(APP_ID)
        # Assert
        assert APP_ID not in manager.smartapps

    @staticmethod
    def test_unregister_no_app_id(manager: SmartAppManager):
        """Test unregister with no SmartApp app id"""
        # Act
        with pytest.raises(ValueError) as e_info:
            manager.unregister(None)
        # Assert
        assert str(e_info.value) == 'smartapp must have an app_id.'

    @staticmethod
    def test_unregister_not_registered(manager: SmartAppManager):
        """Test register with the same app twice"""
        # Act
        with pytest.raises(ValueError) as e_info:
            manager.unregister(APP_ID)
        # Assert
        assert str(e_info.value) == 'smartapp was not previously registered.'

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_config(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("config_init_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_config(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_install(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("install_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_install(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_update(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("update_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_update(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_event(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("event_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_event(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_oauth_callback(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("oauth_callback_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_oauth_callback(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_on_uninstall(manager: SmartAppManager):
        """Tests the config event handler at the manager level."""
        # Arrange
        request = get_fixture("uninstall_request")
        app = manager.register(APP_ID, 'none')
        handler = get_dispatch_handler(app)
        manager.connect_uninstall(handler)
        # Act
        await manager.handle_request(request, None, False)
        # ensure dispatched tasks complete
        await asyncio.gather(*manager.dispatcher.last_sent)
        # Assert
        assert handler.fired



"""Define the configuration module."""

from typing import List

from .const import LIFECYCLE_CONFIG_INIT, LIFECYCLE_CONFIG_PAGE
from .request import Request, Response


class ConfigRequest(Request):
    """Defines a ConfigRequest."""

    def __init__(self, data: dict):
        """Create a new instance of the ConfigRequest."""
        super().__init__(data)
        config = self._config_data_raw = data['configurationData']
        self._installed_app_id = config['installedAppId']
        self._phase = config['phase']
        self._page_id = config['pageId']
        self._previous_page_id = config['previousPageId']

    async def _process(self, app) -> Response:
        if self._phase == LIFECYCLE_CONFIG_INIT:
            resp = self._process_init(app)
        elif self._phase == LIFECYCLE_CONFIG_PAGE:
            resp = self._process_page()
        else:
            raise ValueError("Invalid request configuration phase.")
        return resp

    @staticmethod
    def _process_init(app) -> Response:
        # This is a hard-coded primitive response.
        resp = ConfigInitResponse()
        resp.name = app.name
        resp.config_app_id = app.config_app_id
        resp.description = app.description
        resp.permissions.extend(app.permissions)
        resp.first_page_id = '1'
        return resp

    @staticmethod
    def _process_page() -> Response:
        # This is a hard-coded primitive response.
        resp = ConfigPageResponse()
        resp.page_id = '1'
        resp.name = 'Configuration'
        resp.complete = True
        resp.next_page_id = None
        resp.previous_page_id = None
        return resp

    @property
    def config_data_raw(self) -> dict:
        """Get the raw configuration data."""
        return self._config_data_raw

    @property
    def phase(self) -> str:
        """Get the current phase."""
        return self._phase

    @property
    def page_id(self) -> str:
        """Set the current page id."""
        return self._page_id

    @property
    def previous_page_id(self) -> str:
        """Get the previous page id."""
        return self._previous_page_id


class ConfigInitResponse(Response):
    """Define a configuration init response."""

    def __init__(self):
        """Create a new instance of the ConfigInitResponse."""
        self._name = None
        self._description = None
        self._config_app_id = None
        self._permissions = []
        self._first_page_id = "1"

    def to_data(self) -> dict:
        """Return a data structure representing the response."""
        return {
            "configurationData": {
                "initialize": {
                    "name": self.name,
                    "description": self.description,
                    "id": self.config_app_id,
                    "permissions": self.permissions,
                    "firstPageId": self.first_page_id
                }
            }
        }

    @property
    def name(self) -> str:
        """Get the name of the smartapp."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the smartapp."""
        self._name = value

    @property
    def description(self) -> str:
        """Get the description of the smartapp."""
        return self._description

    @description.setter
    def description(self, value: str):
        """Set the value of the smartapp."""
        self._description = value

    @property
    def config_app_id(self) -> str:
        """Get the id of the smartapp to use in config."""
        return self._config_app_id

    @config_app_id.setter
    def config_app_id(self, value: str):
        """Set the id of the smartapp to use in config."""
        self._config_app_id = value

    @property
    def permissions(self) -> List[str]:
        """Get the permissions the app requires."""
        return self._permissions

    @property
    def first_page_id(self) -> str:
        """Get the id of the first page."""
        return self._first_page_id

    @first_page_id.setter
    def first_page_id(self, value: str):
        """Set the id of the first page."""
        self._first_page_id = value


class ConfigPageResponse(Response):
    """Define a configuration page response."""

    def __init__(self):
        """Create a new instance of the ConfigPageResponse."""
        self._page_id = None
        self._name = None
        self._next_page_id = None
        self._previous_page_id = None
        self._complete = False

    def to_data(self) -> dict:
        """Return a data structure representing the response."""
        return {
            "configurationData": {
                "page": {
                    "pageId": self.page_id,
                    "name": self.name,
                    "nextPageId": self.next_page_id,
                    "previousPageId": self.previous_page_id,
                    "complete": self.complete,
                    "sections": []
                }
            }
        }

    @property
    def page_id(self) -> str:
        """Get the id of the page."""
        return self._page_id

    @page_id.setter
    def page_id(self, value: str):
        """Set the id of the page."""
        self._page_id = value

    @property
    def name(self) -> str:
        """Get the name of the configuration page."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the configuration page."""
        self._name = value

    @property
    def next_page_id(self) -> str:
        """Get the id of the next page."""
        return self._next_page_id

    @next_page_id.setter
    def next_page_id(self, value: str):
        """Set the id of the next page."""
        self._next_page_id = value

    @property
    def previous_page_id(self) -> str:
        """Get the id of the previous page."""
        return self._previous_page_id

    @previous_page_id.setter
    def previous_page_id(self, value: str):
        """Set the id of the previous page."""
        self._previous_page_id = value

    @property
    def complete(self) -> bool:
        """Get whether this is the last config page."""
        return self._complete

    @complete.setter
    def complete(self, value: bool):
        """Set whether this is the last config page."""
        self._complete = value



""""Tests for the ping module."""

from pysmartapp.const import LIFECYCLE_PING
from pysmartapp.ping import PingRequest

from .utilities import get_fixture


class TestPingRequest:
    """Tests for the PingRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('ping_request')
        # Act
        req = PingRequest(data)
        # Assert
        assert req.ping_data_raw == data['pingData']
        assert req.lifecycle == LIFECYCLE_PING
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.ping_challenge == '1a904d57-4fab-4b15-a11e-1c4bfe7cb502'



"""Define the errors module."""


class SignatureVerificationError(Exception):
    """Defines an error for signature verification failures."""

    pass


class SmartAppNotRegisteredError(Exception):
    """Defines an error when a SmartApp isn't registered."""

    _message = "SmartApp handler for installed app '{}' was not found."

    def __init__(self, installed_app_id: str):
        """Create a new instance of the error."""
        Exception.__init__(self, self._message.format(installed_app_id))
        self._installed_app_id = installed_app_id

    @property
    def installed_app_id(self) -> str:
        """Get the installed app id not found."""
        return self._installed_app_id



"""Define the update module."""

from .request import EmptyDataResponse, Request, Response


class UpdateRequest(Request):
    """Define the UpdateRequest class."""

    def __init__(self, data: dict):
        """Create a new instance of the UpdateRequest."""
        super().__init__(data)
        update_data = self._update_data_raw = data['updateData']
        self._init_installed_app(update_data['installedApp'])
        self._auth_token = update_data['authToken']
        self._refresh_token = update_data['refreshToken']

    async def _process(self, app) -> Response:
        resp = EmptyDataResponse('updateData')
        return resp

    @property
    def update_data_raw(self) -> dict:
        """Get the raw update data."""
        return self._update_data_raw

    @property
    def auth_token(self):
        """Get the auth token."""
        return self._auth_token

    @property
    def refresh_token(self):
        """Get the refresh token."""
        return self._refresh_token



"""Define the pysmartapp package."""

from pysmartapp.config import (
    ConfigInitResponse, ConfigPageResponse, ConfigRequest)
from pysmartapp.const import __title__, __version__  # noqa
from pysmartapp.dispatch import Dispatcher
from pysmartapp.errors import (
    SignatureVerificationError, SmartAppNotRegisteredError)
from pysmartapp.event import Event, EventRequest
from pysmartapp.install import InstallRequest
from pysmartapp.oauthcallback import OAuthCallbackRequest
from pysmartapp.ping import PingRequest, PingResponse
from pysmartapp.request import EmptyDataResponse, Request, Response
from pysmartapp.smartapp import SmartApp, SmartAppManager
from pysmartapp.uninstall import UninstallRequest
from pysmartapp.update import UpdateRequest

__all__ = [
    # config
    'ConfigInitResponse',
    'ConfigPageResponse',
    'ConfigRequest',
    # dispatch
    'Dispatcher',
    # errors
    'SignatureVerificationError',
    'SmartAppNotRegisteredError',
    # event
    'Event',
    'EventRequest',
    # install
    'InstallRequest',
    # oauthcallback
    'OAuthCallbackRequest',
    # ping
    'PingRequest',
    'PingResponse',
    # request
    'EmptyDataResponse',
    'Request',
    'Response',
    # smartapp
    'SmartApp',
    'SmartAppManager',
    # unisntall
    'UninstallRequest',
    'UpdateRequest'
]



"""Define tests for the Dispatch module."""

import asyncio

import pytest

from pysmartapp.dispatch import Dispatcher


class TestDispatcher:
    """Define tests for the dispatcher class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_connect(handler):
        """Tests the connect function."""
        # Arrange
        dispatcher = Dispatcher()
        # Act
        dispatcher.connect('TEST', handler)
        # Assert
        assert handler in dispatcher.signals['TEST']

    @staticmethod
    @pytest.mark.asyncio
    async def test_disconnect(handler):
        """Tests the disconnect function."""
        # Arrange
        dispatcher = Dispatcher()
        disconnect = dispatcher.connect('TEST', handler)
        # Act
        disconnect()
        # Assert
        assert handler not in dispatcher.signals['TEST']

    @staticmethod
    @pytest.mark.asyncio
    async def test_already_disconnected(handler):
        """Tests that disconnect can be called more than once."""
        # Arrange
        dispatcher = Dispatcher()
        disconnect = dispatcher.connect('TEST', handler)
        disconnect()
        # Act
        disconnect()
        # Assert
        assert handler not in dispatcher.signals['TEST']

    @staticmethod
    @pytest.mark.asyncio
    async def test_send_async_handler(async_handler):
        """Tests sending to async handlers."""
        # Arrange
        dispatcher = Dispatcher()
        dispatcher.connect('TEST', async_handler)
        # Act
        await asyncio.gather(*dispatcher.send('TEST'))
        # Assert
        assert async_handler.fired

    @staticmethod
    @pytest.mark.asyncio
    async def test_send(handler):
        """Tests sending to async handlers."""
        # Arrange
        dispatcher = Dispatcher()
        dispatcher.connect('TEST', handler)
        args = object()
        # Act
        await asyncio.gather(*dispatcher.send('TEST', args))
        # Assert
        assert handler.fired
        assert handler.args[0] == args

    @staticmethod
    @pytest.mark.asyncio
    async def test_custom_connect_and_send(handler):
        """Tests using the custom connect and send implementations."""
        # Arrange
        test_signal = 'PREFIX_TEST'
        stored_target = None

        def connect(signal, target):
            assert signal == test_signal
            nonlocal stored_target
            stored_target = target

            def disconnect():
                nonlocal stored_target
                stored_target = None
            return disconnect

        def send(signal, *args):
            assert signal == test_signal
            stored_target(*args)  # pylint:disable=not-callable

        dispatcher = Dispatcher(connect=connect, send=send,
                                signal_prefix='PREFIX_')

        # Act
        dispatcher.connect('TEST', handler)
        dispatcher.send('TEST')
        # Assert
        assert handler.fired



"""Define the request module."""

from httpsig.verify import HeaderVerifier

from .errors import SignatureVerificationError


class Response:
    """Defines a response."""

    def to_data(self) -> dict:
        """Create a data structure for the response."""
        raise NotImplementedError


class EmptyDataResponse(Response):
    """Defines a response with an empty data structure."""

    def __init__(self, name: str):
        """Create a new instance of the EmptyDataResponse class."""
        self._name = name

    def to_data(self) -> dict:
        """Return a data structure representing this request."""
        return {self.name: {}}

    @property
    def name(self) -> str:
        """Get the name of the empty data tag."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the empty data tag."""
        self._name = value


class Request:
    """Defines a request to process."""

    def __init__(self, data: dict):
        """Create a new instance of the Request class."""
        self._lifecycle = data['lifecycle']
        self._execution_id = data['executionId']
        self._locale = data['locale']
        self._version = data['version']
        self._installed_app_id = None
        self._location_id = None
        self._installed_app_config = {}
        self._settings = data.get('settings', {})
        self._supports_validation = True

    def _init_installed_app(self, installed_app):
        self._installed_app_id = installed_app['installedAppId']
        self._location_id = installed_app['locationId']
        self._installed_app_config = installed_app['config']

    async def process(self, app, headers: list = None,
                      validate_signature: bool = True) -> Response:
        """Process the request with the SmartApp."""
        if validate_signature and self._supports_validation:
            try:
                verifier = HeaderVerifier(
                    headers=headers, secret=app.public_key, method='POST',
                    path=app.path)
                result = verifier.verify()
            except Exception as ex:
                raise SignatureVerificationError from ex
            if not result:
                raise SignatureVerificationError
        response = await self._process(app)
        app.dispatcher.send(self.lifecycle, self, response, app)
        return response

    async def _process(self, app) -> Response:
        raise NotImplementedError

    @property
    def lifecycle(self) -> str:
        """Get the lifecycle of the request."""
        return self._lifecycle

    @property
    def execution_id(self) -> str:
        """Get the execution id of the request."""
        return self._execution_id

    @property
    def locale(self) -> str:
        """Get the locale of the request."""
        return self._locale

    @property
    def version(self) -> str:
        """Get the version of the request."""
        return self._version

    @property
    def installed_app_id(self) -> str:
        """Get the installed app id the request is for."""
        return self._installed_app_id

    @property
    def location_id(self) -> str:
        """Get the installed app location id."""
        return self._location_id

    @property
    def installed_app_config(self) -> dict:
        """Get the installed app configuration."""
        return self._installed_app_config

    @property
    def settings(self):
        """Get the settings associated with the request."""
        return self._settings



"""Testing utilities."""
import json


def get_fixture(file: str, ext: str = 'json'):
    """Load a fixtures file."""
    file_name = "tests/fixtures/{file}.{ext}".format(file=file, ext=ext)
    with open(file_name) as open_file:
        if ext == 'json':
            return json.load(open_file)
        return open_file.read()


def get_dispatch_handler(smartapp):
    """Get a handler to mock in the dispatcher."""
    async def handler(req, resp, app):
        handler.fired = True
        assert app == smartapp
    handler.fired = False
    return handler



"""Define the event module."""

from typing import Sequence

from .const import EVENT_TYPE_DEVICE, EVENT_TYPE_TIMER
from .request import EmptyDataResponse, Request, Response


class Event:
    """Define an event."""

    def __init__(self, data: dict):
        """Create a new instance of the event class."""
        self._event_type = data['eventType']
        self._subscription_name = None
        self._event_id = None
        self._location_id = None
        self._device_id = None
        self._component_id = None
        self._capability = None
        self._attribute = None
        self._value = None
        self._state_change = None
        if self._event_type == EVENT_TYPE_DEVICE:
            device_event = data['deviceEvent']
            self._subscription_name = device_event['subscriptionName']
            self._event_id = device_event['eventId']
            self._location_id = device_event['locationId']
            self._device_id = device_event['deviceId']
            self._component_id = device_event['componentId']
            self._capability = device_event['capability']
            self._attribute = device_event['attribute']
            self._value = device_event['value']
            self._state_change = device_event['stateChange']
        self._timer_name = None
        self._timer_type = None
        self._timer_time = None
        self._timer_expression = None
        if self._event_type == EVENT_TYPE_TIMER:
            timer_event = data['timerEvent']
            self._event_id = timer_event['eventId']
            self._timer_name = timer_event['name']
            self._timer_type = timer_event['type']
            self._timer_time = timer_event['time']
            self._timer_expression = timer_event['expression']

    @property
    def event_type(self) -> str:
        """Get the type of the event."""
        return self._event_type

    @property
    def subscription_name(self) -> str:
        """Get the subscription name."""
        return self._subscription_name

    @property
    def event_id(self) -> str:
        """Get the event id."""
        return self._event_id

    @property
    def location_id(self) -> str:
        """Get the location id."""
        return self._location_id

    @property
    def device_id(self) -> str:
        """Get the device id."""
        return self._device_id

    @property
    def component_id(self) -> str:
        """Get the component id."""
        return self._component_id

    @property
    def capability(self) -> str:
        """Get the capability."""
        return self._capability

    @property
    def attribute(self) -> str:
        """Get the attribute."""
        return self._attribute

    @property
    def value(self):
        """Get the value."""
        return self._value

    @property
    def state_change(self) -> bool:
        """Get whether this is a new state change."""
        return self._state_change

    @property
    def timer_name(self) -> str:
        """Get the name of the timer schedule."""
        return self._timer_name

    @property
    def timer_type(self) -> str:
        """Get the type of time."""
        return self._timer_type

    @property
    def timer_time(self) -> str:
        """Get the time the timer fired."""
        return self._timer_time

    @property
    def timer_expression(self) -> str:
        """Get the timer firing expression."""
        return self._timer_expression


class EventRequest(Request):
    """Define the EventRequest class."""

    def __init__(self, data: dict):
        """Create a new instance of the EventRequest."""
        super().__init__(data)
        event_data = self._event_data_raw = data['eventData']
        self._auth_token = event_data['authToken']
        self._init_installed_app(event_data['installedApp'])
        self._events = [Event(item) for item in event_data['events']]

    async def _process(self, app) -> Response:
        resp = EmptyDataResponse('eventData')
        return resp

    @property
    def event_data_raw(self) -> dict:
        """Get the raw event data."""
        return self._event_data_raw

    @property
    def auth_token(self):
        """Get the auth token."""
        return self._auth_token

    @property
    def events(self) -> Sequence[Event]:
        """Get the events."""
        return self._events



"""Define a SmartApp."""

import logging
from typing import Any, Callable, Dict, List

from .const import (
    LIFECYCLE_CONFIG, LIFECYCLE_EVENT, LIFECYCLE_INSTALL,
    LIFECYCLE_OAUTH_CALLBACK, LIFECYCLE_PING, LIFECYCLE_UNINSTALL,
    LIFECYCLE_UPDATE, SETTINGS_APP_ID)
from .dispatch import Dispatcher
from .errors import SmartAppNotRegisteredError
from .utilities import create_request

_LOGGER = logging.getLogger(__name__)


class SmartAppBase:
    """Define common functionality for the SmartApp and SmartAppManager."""

    def __init__(self, *, path: str = '/', dispatcher: Dispatcher = None):
        """Initialize a new instance of the smartapp."""
        self._dispatcher = dispatcher or Dispatcher()
        self._path = path

    def connect_ping(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the ping signal."""
        return self._dispatcher.connect(LIFECYCLE_PING, target)

    def connect_config(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the config signal."""
        return self._dispatcher.connect(LIFECYCLE_CONFIG, target)

    def connect_install(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the install signal."""
        return self._dispatcher.connect(LIFECYCLE_INSTALL, target)

    def connect_update(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the update signal."""
        return self._dispatcher.connect(LIFECYCLE_UPDATE, target)

    def connect_event(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the event signal."""
        return self._dispatcher.connect(LIFECYCLE_EVENT, target)

    def connect_oauth_callback(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the oauth callback signal."""
        return self._dispatcher.connect(LIFECYCLE_OAUTH_CALLBACK, target)

    def connect_uninstall(self, target: Callable[..., Any]) \
            -> Callable[[], None]:
        """Connect a target to the oauth callback signal."""
        return self._dispatcher.connect(LIFECYCLE_UNINSTALL, target)

    @property
    def path(self) -> str:
        """Get the path the SmartApp is installed at."""
        return self._path

    @property
    def dispatcher(self):
        """Get the dispatcher used to connect and send notifications."""
        return self._dispatcher


class SmartApp(SmartAppBase):
    """Define the SmartApp class."""

    def __init__(self, *, path: str = '/', public_key=None,
                 dispatcher: Dispatcher = None):
        """Initialize the SmartApp class."""
        super().__init__(path=path, dispatcher=dispatcher)
        self._app_id = None
        self._config_app_id = 'app'
        self._description = None
        self._name = None
        self._permissions = []
        self._public_key = public_key

    async def handle_request(self, data: dict, headers: dict = None,
                             validate_signature: bool = True) -> dict:
        """Process a lifecycle event."""
        req = create_request(data)
        resp = await req.process(self, headers, validate_signature)

        if req.installed_app_id:
            _LOGGER.debug("%s: %s received for installed app %s.",
                          req.execution_id, req.lifecycle,
                          req.installed_app_id)
        else:
            _LOGGER.debug("%s: %s received.",
                          req.execution_id, req.lifecycle)

        return resp.to_data()

    @property
    def app_id(self):
        """Get the unique id of the SmartApp."""
        return self._app_id

    @app_id.setter
    def app_id(self, value: str):
        """Set the app id of the SmartApp."""
        self._app_id = value

    @property
    def name(self) -> str:
        """Get the name of the SmartApp using during config."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the SmartApp used during config."""
        self._name = value

    @property
    def description(self) -> str:
        """Get the description of the SmartApp used during config."""
        return self._description

    @description.setter
    def description(self, value: str):
        """Set the description of the SmartApp used during config."""
        self._description = value

    @property
    def config_app_id(self):
        """Get the id for the SmartApp used during config."""
        return self._config_app_id

    @config_app_id.setter
    def config_app_id(self, value: str):
        """Set the id of the SmartApp used during config."""
        self._config_app_id = value

    @property
    def permissions(self) -> List[str]:
        """Get the permissions of the SmartApp used during config."""
        return self._permissions

    @property
    def public_key(self):
        """Get the public key of the SmartApp used to verify events."""
        return self._public_key


class SmartAppManager(SmartAppBase):
    """Service to support multiple SmartApps at the same end-point."""

    def __init__(self, path: str, *, dispatcher: Dispatcher = None):
        """Create a new instance of the manager."""
        super().__init__(path=path, dispatcher=dispatcher)
        self._smartapps = {}

    async def handle_request(self, data: dict, headers: dict = None,
                             validate_signature: bool = True) -> dict:
        """Process a lifecycle event."""
        req = create_request(data)
        # Always process ping lifecycle events.
        if req.lifecycle == LIFECYCLE_PING:
            resp = await req.process(self, headers, validate_signature)
        else:
            app_id = req.settings.get(SETTINGS_APP_ID)
            if not app_id:
                raise SmartAppNotRegisteredError(req.installed_app_id)
            smartapp = self._smartapps.get(app_id)
            if not smartapp:
                raise SmartAppNotRegisteredError(req.installed_app_id)
            resp = await req.process(smartapp, headers, validate_signature)

        if req.installed_app_id:
            _LOGGER.debug("%s: %s received for installed app %s.",
                          req.execution_id, req.lifecycle,
                          req.installed_app_id)
        else:
            _LOGGER.debug("%s: %s received.",
                          req.execution_id, req.lifecycle)
        return resp.to_data()

    def register(self, app_id: str, public_key: str) -> SmartApp:
        """Create a new SmartApp for the end-point."""
        if app_id is None:
            raise ValueError('smartapp must have an app_id.')
        if app_id in self._smartapps:
            raise ValueError('smartapp already registered.')
        smartapp = SmartApp(
            path=self._path,
            public_key=public_key,
            dispatcher=self._dispatcher
        )
        smartapp.app_id = app_id
        self._smartapps[smartapp.app_id] = smartapp
        return smartapp

    def unregister(self, app_id: str):
        """Unregister the specified installed app id."""
        if app_id is None:
            raise ValueError('smartapp must have an app_id.')
        if app_id not in self._smartapps:
            raise ValueError('smartapp was not previously registered.')
        self._smartapps.pop(app_id, None)

    @property
    def smartapps(self) -> Dict[str, SmartApp]:
        """Get registered SmartApps."""
        return self._smartapps



""""Tests for the install module."""

from pysmartapp.const import LIFECYCLE_INSTALL
from pysmartapp.install import InstallRequest

from .utilities import get_fixture


class TestInstallRequest:
    """Tests for the InstallRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('install_request')
        # Act
        req = InstallRequest(data)
        # Assert
        assert req.install_data_raw == data['installData']
        assert req.lifecycle == LIFECYCLE_INSTALL
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == '8a0dcdc9-1ab4-4c60-9de7-cb78f59a1121'
        assert req.location_id == 'e675a3d9-2499-406c-86dc-8a492a886494'
        assert req.installed_app_config ==\
            data['installData']['installedApp']['config']
        assert req.settings == data['settings']
        assert req.auth_token == '580aff1f-f0f1-44e0-94d4-e68bf9c2e768'
        assert req.refresh_token == 'ad58374e-9d6a-4457-8488-a05aa8337ab3'



""""Tests for the event module."""

from pysmartapp.const import (
    EVENT_TYPE_DEVICE, EVENT_TYPE_TIMER, LIFECYCLE_EVENT)
from pysmartapp.event import Event, EventRequest

from .utilities import get_fixture


class TestEventRequest:
    """Tests for the EventRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('event_request')
        # Act
        req = EventRequest(data)
        # Assert
        assert req.event_data_raw == data['eventData']
        assert req.lifecycle == LIFECYCLE_EVENT
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == '8a0dcdc9-1ab4-4c60-9de7-cb78f59a1121'
        assert req.location_id == 'e675a3d9-2499-406c-86dc-8a492a886494'
        assert req.installed_app_config == {}
        assert req.settings == data['settings']
        assert req.auth_token == 'f01894ce-013a-434a-b51e-f82126fd72e4'
        assert len(req.events) == 2


class TestEvent:
    """Tests for the Event class."""

    @staticmethod
    def test_init_device_event():
        """Tests the init method."""
        # Arrange
        data = get_fixture('event_request')['eventData']['events'][0]
        # Act
        evt = Event(data)
        # Assert
        assert evt.event_type == EVENT_TYPE_DEVICE
        assert evt.subscription_name == 'motion_sensors'
        assert evt.event_id == '736e3903-001c-4d40-b408-ff40d162a06b'
        assert evt.location_id == '499e28ba-b33b-49c9-a5a1-cce40e41f8a6'
        assert evt.device_id == '6f5ea629-4c05-4a90-a244-cc129b0a80c3'
        assert evt.component_id == 'main'
        assert evt.capability == 'motionSensor'
        assert evt.attribute == 'motion'
        assert evt.value == 'active'
        assert evt.state_change

    @staticmethod
    def test_init_timer_event():
        """Tests the init method."""
        # Arrange
        data = get_fixture('event_request')['eventData']['events'][1]
        # Act
        evt = Event(data)
        # Assert
        assert evt.event_type == EVENT_TYPE_TIMER
        assert evt.event_id == 'string'
        assert evt.timer_name == 'lights_off_timeout'
        assert evt.timer_type == 'CRON'
        assert evt.timer_time == '2017-09-13T04:18:12.469Z'
        assert evt.timer_expression == 'string'



"""Define the utilities class."""

from .config import ConfigRequest
from .const import (
    LIFECYCLE_CONFIG, LIFECYCLE_EVENT, LIFECYCLE_INSTALL,
    LIFECYCLE_OAUTH_CALLBACK, LIFECYCLE_PING, LIFECYCLE_UNINSTALL,
    LIFECYCLE_UPDATE)
from .event import EventRequest
from .install import InstallRequest
from .oauthcallback import OAuthCallbackRequest
from .ping import PingRequest
from .uninstall import UninstallRequest
from .update import UpdateRequest


def create_request(data: dict):
    """Create a request from the given dictionary and headers."""
    lifecycle = data['lifecycle']
    if lifecycle == LIFECYCLE_PING:
        return PingRequest(data)
    if lifecycle == LIFECYCLE_CONFIG:
        return ConfigRequest(data)
    if lifecycle == LIFECYCLE_INSTALL:
        return InstallRequest(data)
    if lifecycle == LIFECYCLE_UPDATE:
        return UpdateRequest(data)
    if lifecycle == LIFECYCLE_EVENT:
        return EventRequest(data)
    if lifecycle == LIFECYCLE_OAUTH_CALLBACK:
        return OAuthCallbackRequest(data)
    if lifecycle == LIFECYCLE_UNINSTALL:
        return UninstallRequest(data)
    raise ValueError('The specified lifecycle event was not recognized.')



"""Define the oauthcallback module."""

from .request import EmptyDataResponse, Request, Response


class OAuthCallbackRequest(Request):
    """Define the OAuthCallbackRequest class."""

    def __init__(self, data: dict):
        """Create a new instance of the OAuthCallbackRequest."""
        super().__init__(data)
        callback_data = self._oauth_callback_data_raw = \
            data['oAuthCallbackData']
        self._installed_app_id = callback_data['installedAppId']
        self._url_path = callback_data['urlPath']

    async def _process(self, app) -> Response:
        resp = EmptyDataResponse('oAuthCallbackData')
        return resp

    @property
    def oauth_callback_data_raw(self) -> dict:
        """Get the raw OAuth Callback data."""
        return self._oauth_callback_data_raw

    @property
    def url_path(self) -> str:
        """Get the url path of the OAuth callback."""
        return self._url_path



"""Define constants used in the SmartApp."""

__title__ = "pysmartapp"
__version__ = "0.3.0"

LIFECYCLE_PING = 'PING'
LIFECYCLE_CONFIG = 'CONFIGURATION'
LIFECYCLE_CONFIG_INIT = 'INITIALIZE'
LIFECYCLE_CONFIG_PAGE = 'PAGE'
LIFECYCLE_INSTALL = 'INSTALL'
LIFECYCLE_UPDATE = 'UPDATE'
LIFECYCLE_EVENT = 'EVENT'
LIFECYCLE_OAUTH_CALLBACK = 'OAUTH_CALLBACK'
LIFECYCLE_UNINSTALL = 'UNINSTALL'
EVENT_TYPE_DEVICE = 'DEVICE_EVENT'
EVENT_TYPE_TIMER = 'TIMER_EVENT'
SETTINGS_APP_ID = 'appId'



""""Tests for the update module."""

from pysmartapp.const import LIFECYCLE_UPDATE
from pysmartapp.update import UpdateRequest

from .utilities import get_fixture


class TestUpdateRequest:
    """Tests for the UpdateRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('update_request')
        # Act
        req = UpdateRequest(data)
        # Assert
        assert req.update_data_raw == data['updateData']
        assert req.lifecycle == LIFECYCLE_UPDATE
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == '8a0dcdc9-1ab4-4c60-9de7-cb78f59a1121'
        assert req.location_id == 'e675a3d9-2499-406c-86dc-8a492a886494'
        assert req.installed_app_config == \
            data['updateData']['installedApp']['config']
        assert req.settings == data['settings']
        assert req.auth_token == '4ebd8d9f-53b0-483f-a989-4bde30ca83c0'
        assert req.refresh_token == '6e3bbf5f-b68d-4250-bbc9-f7151016a77f'



"""Tests for the utilities module."""

import pytest

from pysmartapp import utilities


def test_create_request_invalid():
    """Tests the create_request method."""
    # Arrange
    data = {'lifecycle': 'UNKNOWN'}
    # Act/Assert
    with pytest.raises(ValueError):
        utilities.create_request(data)



"""Define the install module."""

from .request import EmptyDataResponse, Request, Response


class InstallRequest(Request):
    """Define the InstallRequest class."""

    def __init__(self, data: dict):
        """Create a new instance of the InstallRequest."""
        super().__init__(data)
        install_data = self._install_data_raw = data['installData']
        self._init_installed_app(install_data['installedApp'])
        self._auth_token = install_data['authToken']
        self._refresh_token = install_data['refreshToken']

    async def _process(self, app) -> Response:
        resp = EmptyDataResponse('installData')
        return resp

    @property
    def install_data_raw(self) -> dict:
        """Get the raw installation data."""
        return self._install_data_raw

    @property
    def auth_token(self):
        """Get the auth token."""
        return self._auth_token

    @property
    def refresh_token(self):
        """Get the refresh token."""
        return self._refresh_token
